CON

  _clkmode      = xtal1 + pll8x 'Set for Propeller Backpack. Edit for 5 MHz boards.
  _xinfreq      = 10_000_000

  LPIN          = 0             'Pin numbers for Ping)))s. Must be contiguous and in this order.
  MPIN          = 1
  RPIN          = 2

  SEPX          = 105           'Horizontal separation of Ping)))s in array (inch * 100).
  SEPZ          = 100           'Vertical separation of transmit and recive transducers (inch * 100).

  RANGE         = 20            'Search range (plus and minus) to search for best convergence of locations.
  INCR          = 5             'Search increment.

OBJ

  pst :         "Parallax Serial Terminal"
  num :         "Simple_Numbers_Plus"

VAR

  long x, y, xavg, yavg, xlm, xmr, xlr, dl, dm, dr, minerr, err, l, m, r, bestdm

PUB  start | t0, t1, i, str, ix, iy

  'Display the grid.

  pst.start(38400)
  pst.clear
  repeat i from 0 to 36
    pst.position(0, i + 2)
    if (i // 3)
      str := string("|     ")
      pst.str(string("   "))
    else
      str := string("+-----")
      pst.str(num.decn(12 - i / 3, 3))
    repeat 10
      pst.str(str)
    pst.char(byte[str])
  pst.char(13)
  pst.str(string("  -5    -4    -3    -2    -1     0     1     2     3     4     5"))

  'Initialize frqx and Ping))) pins.
  
  frqa := frqb := 1
  outa[LPIN .. RPIN]~~

  'Main program loop. 
  
  repeat
    waitcnt(cnt + clkfreq / 30)                         'Wait for previous echoes to die.
    phsa~                                               'Clear the counters.
    phsb~
    dira[0..2]~~                                        'Trigger all three Ping)))s
    dira[0..2]~
    ctra := ctrb := constant(%01000 << 26 | MPIN)       'Set L & R counters to count when middle Ping))) output is high.
    waitpne(0, constant(1 << MPIN), 0)                  'Wait for middle output to go high.
    t0 := cnt                                           'Get initial time for middle Ping))).
    waitpeq(constant(%111 << LPIN), constant(%111 << LPIN), 0) 'Wait for all outputs to go high.
    ctra := constant(%01000 << 26 | LPIN)               'Change input pin for ctra to left Ping))).
    ctrb := constant(%01000 << 26 | RPIN)               'Change input pin for ctrb to right Ping))).
    waitpeq(0, constant(1 << MPIN), 0)                  'Wait for middle output to go low.
    t1 := cnt                                           'Capture the time for it.
    waitpeq(0, %111 << LPIN, 0)                         'Wait for all outputs to go low.
    dm := ((t1 - t0) * 1351 / (clkfreq / 1000))         'Read, compute, and scale the times to hundredths of an inch.
    dl := phsa * 1351 / (clkfreq / 1000)
    dr := phsb * 1351 / (clkfreq / 1000)
    dm >>= 1
    dl -= dm
    dr -= dm
    repeat i from 0 to 2
      dl[i] := ^^(dl[i] * dl[i] - constant(SEPZ * SEPZ / 4))

    'Replace the previous marker on the screen with the background.
            
    pst.position(ix, iy)
    if ((ix - 3) // 6)
      if ((iy - 2) // 3)
        pst.char(" ")
      else
        pst.char("-")
    else
      if ((iy - 2) // 3)
        pst.char("|")
      else
        pst.char("+")

    'Minimize the position error.
        
    minerr := $7fff_ffff                                'Initialize minimum error.
    repeat l from dl - RANGE to dl + RANGE step INCR    'Increment around all three distances, and compute X values.
      repeat m from dm - RANGE to dm + RANGE step INCR
        xlm := (l * l - m * m - constant(SEPX * SEPX)) / constant(2 * SEPX)
        repeat r from dr - RANGE to dr + RANGE step INCR
          xmr := (m * m - r * r + constant(SEPX * SEPX)) / constant(2 * SEPX)
          xlr := (l * l - r * r) / constant(4 * SEPX)
          err := (xlm - xmr) * (xlm - xmr) + (xmr - xlr) * (xmr - xlr) + (xlm - xlr) * (xlm - xlr) 'Error is sum-of-squared differences.
          if (err < minerr)                             'Is error less than minimum seen so far?
            minerr := err                               '  Yes: New minerr.
            x := (xlm + xmr + xlr) / 3                  '       New X.
            bestdm := m                                 '       New best middle distance for computing Y.                    
    y := ^^(bestdm * bestdm - x * x)                    'Compute Y.
    if (yavg)                                           'Started averaging position yet?
      xavg := (xavg * 9 + x) / 10                       '  Yes: Do running average. 
      yavg := (yavg * 9 + y) / 10
    else
      xavg := x                                         '  No:  Initialize with current value.
      yavg := y
    ix := xavg * 6 / 100 + 33 #> 3 <# 63                'Compute new screen position of marker.
    iy := 38 - (yavg * 3 + 50) / 100 #> 2 <# 38
    pst.position(ix, iy)                                'Plot it.
    pst.char("O")

    'Display X, Y, and Error.

    pst.position(16, 0)
    pst.str(string("X: "))
    pst.str(num.decf(xavg, 2))                          'Display X as two decimal places.
    pst.char(34)
    pst.clearend
    pst.position(28, 0)
    pst.str(string("Y: "))                              'Display Y as two decimal places.    
    pst.str(num.decf(yavg, 2))
    pst.char(34)
    pst.position(40, 0)
    pst.str(string("Err: "))    
    pst.str(num.decf(^^(minerr / 3), 2))                'Display RMS error as two decimal places.
    pst.char(34)
    pst.char(pst#CE)
          